from typing import Optional

from uiwiz.element import Element
from uiwiz.elements.extensions.on_event import OnEvent


class Dropdown(OnEvent):
    root_class: str = "select "
    root_size: str = "select-{size}"
    _classes: str = "select-bordered w-full max-w-xs"

    def __init__(self, name: str, items: list[str], placeholder: Optional[str] = None) -> None:
        super().__init__("select")
        self.attributes["name"] = name
        self.classes()
        self.size(self._size)
        with self:
            if placeholder and placeholder not in items:
                Element("option disabled selected", content=placeholder)
            for v in items:
                e = Element("option", content=v)
                e.value = v
                if placeholder == v:
                    e.attributes["selected"] = "selected"



from uiwiz.element import Element


class Html(Element):
    def __init__(self, content) -> None:
        super().__init__()
        self.__content__ = content

    @property
    def content(self) -> str:
        return self.__content__

    @content.setter
    def content(self, content):
        self.__content__ = content



from typing import Optional

from uiwiz.element import Element
from uiwiz.elements.extensions.on_event import OnEvent


class Input(OnEvent):
    root_class: str = "input "
    root_size: str = "input-{size}"
    _classes: str = "input-bordered w-full"

    def __init__(
        self,
        name: Optional[str] = None,
        value: Optional[str] = None,
        placeholder: Optional[str] = None,
    ) -> None:
        """Input

        This element is used for input data

        :param name: Used in the attributes name sent back in json
        :param value: The default value for the input
        :param placeholder: Text to show before input
        """
        super().__init__("input")
        self.classes(Input._classes)
        self.attributes["name"] = name
        self.placeholder = placeholder
        if value:
            self.value = value
        self.attributes["autocomplete"] = "off"

    @property
    def placeholder(self) -> Optional[str]:
        return self.attributes.get("placeholder")

    @placeholder.setter
    def placeholder(self, value: Optional[str]):
        if value:
            self.attributes["placeholder"] = value
        return self

    def set_placeholder(self, value: str) -> "Input":
        self.placeholder = value
        return self

    def set_floating_label(self, label: str) -> "Input":
        if self.placeholder is None:
            raise ValueError("Placeholder must be set before floating label")

        self.parent_element.children.remove(self)
        with Element("label").classes("floating-label") as container:
            self.label_text = Element("span", content=label)
            container.children.append(self)
            self.parent_element = container

        return self



import hashlib
import inspect
from functools import lru_cache
from pathlib import Path
from typing import Callable, Dict, Optional

resources: Dict[str, Path] = {}
page_map: Dict[Callable, str] = {}


@lru_cache(maxsize=None)
def hash_function_extended(func):
    """
    This was an interesting problem. I needed to hash the function
    to be able to store the route in a dictionary. Nothing special
    but the reason for using the source code has to do with
    uvicorn and multiple workers. The function object is not the
    same in different workers, so I had to use the source code
    to generate a hash that would be the same in all workers.
    """
    source = inspect.getsource(func) + func.__name__
    return hashlib.sha256(source.encode("utf-8")).hexdigest()


def register_resource(key: str, resource: Path):
    resources[key] = resource


def register_path(key: str, func: Callable):
    page_map[hash_function_extended(func)] = key


def fetch_route(func: Callable) -> Optional[str]:
    return page_map.get(hash_function_extended(func))


def reset_resources():
    resources.clear()
    page_map.clear()


def route_exists(path: str) -> bool:
    return path in list(page_map.values())



import logging
from html import escape
from mimetypes import guess_type
from pathlib import Path
from typing import Optional, Union

from fastapi import FastAPI, Response
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import HTMLResponse, JSONResponse
from fastapi.routing import APIRoute
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from jinja2 import Template
from starlette.requests import Request

from uiwiz.asgi_request_middleware import AsgiRequestMiddleware
from uiwiz.frame import Frame
from uiwiz.page_route import PageRouter
from uiwiz.shared import register_path, resources
from uiwiz.static_middleware import AsgiTtlMiddleware
from uiwiz.version import __version__

logger = logging.getLogger("uiwiz")
logger.addHandler(logging.NullHandler())


class CustomList(list):
    def append(self, route: APIRoute):
        if getattr(route, "endpoint", None):
            register_path(route.path, route.endpoint)
        return super().append(route)


class UiwizApp(FastAPI):
    def __init__(
        self,
        toast_delay: int = 2500,
        error_classes: str = "alert alert-error",
        cache_age: int = 14400,
        theme: Optional[str] = None,
        auth_header: Optional[str] = None,
        title: Optional[str] = "UiWiz",
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.router.routes = CustomList()
        self.toast_delay = toast_delay
        self.error_classes = error_classes
        if theme:
            self.theme = f"data-theme={theme}"
        else:
            self.theme = theme
        self.auth_header = auth_header
        self.title = title
        self.templates = Jinja2Templates(Path(__file__).parent / "templates")
        self.add_static_files(f"/_static/{__version__}/", Path(__file__).parent / "static")

        self.add_middleware(AsgiRequestMiddleware)
        self.add_middleware(GZipMiddleware)
        self.add_middleware(AsgiTtlMiddleware, cache_age=cache_age)
        self.extensions: dict[str, Path] = {}
        self.app_paths: dict[str, Path] = {}

        @self.get(f"/_uiwiz/{__version__}/default.js", include_in_schema=False)
        def return_default_js(request: Request, response: Response):
            response.headers["cache-control"] = f"max-age={cache_age}"
            response.headers["x-uiwiz-content"] = "assets"
            return self.render(request, response, template_name="default.js", media_type="application/javascript")

        @self.exception_handler(RequestValidationError)
        async def handle_validation_error(request: Request, exc: RequestValidationError):
            fields_with_errors = [item.get("loc")[1] for item in exc.errors() if item.get("loc")[1] in exc.body]
            ok_fields = [item for item in exc.body.keys() if item not in fields_with_errors]
            message = " <br> ".join(
                [f"{item.get('loc')[1]}: {item.get('msg')}" for item in exc.errors() if item.get("loc")[1] in exc.body]
            )
            return JSONResponse(
                status_code=422,
                content=jsonable_encoder(
                    {
                        "detail": exc.errors(),
                        "fieldErrors": fields_with_errors,
                        "fieldOk": ok_fields,
                        "message": message,
                    }
                ),
            )

        @self.get("/_static/extension/{__version__}/{extension}/{filename}", include_in_schema=False)
        def get_extension(extension: str, filename: str):
            resource_key = f"{extension}/{filename}"
            if resource_key not in resources:
                return Response(status_code=404)

            with open(resources[resource_key], encoding="utf-8") as f:
                content = f.read()

            content_type, _ = guess_type(resource_key)
            return Response(content, media_type=content_type)

    def __get_title__(self, frame: Frame, route_title: Optional[str] = None) -> str:
        title = self.title

        if route_title:
            title = route_title

        if dynamic_title := frame.title:
            title = dynamic_title

        return title

    def add_static_files(self, url_path: str, local_directory: Union[str, Path]) -> None:
        self.mount(url_path, StaticFiles(directory=str(local_directory)))

    def page(
        self, path: str, *args, title: Optional[str] = None, favicon: Optional[str] = None, **kwargs
    ) -> PageRouter:
        return PageRouter().page(path, *args, title=title, favicon=favicon, router=self.router, **kwargs)

    def ui(self, path: str, *args, include_js: bool = True, include_css: bool = True, **kwargs) -> PageRouter:
        return PageRouter().ui(path=path, include_js=include_js, include_css=include_css, router=self.router, **kwargs)

    def render(
        self,
        request: Request,
        response: Response,
        title: Optional[str] = None,
        status_code: int = 200,
        template_name: str = "default.html",
        media_type: str = "text/html",
        root_overflow: str = "overflow-y: scroll",
    ):
        frame = Frame.get_stack()
        ext_js, ext_css = frame.render_ext()

        theme = self.theme
        if cookie_theme := request.cookies.get("data-theme"):
            theme = f"data-theme={escape(cookie_theme)}"

        root_overflow = f'style="{root_overflow}"'
        standard_headers = {"cache-control": "no-store", "x-uiwiz-content": "page"}
        default_page: Template = self.templates.get_template(template_name)

        if response:
            for key, value in response.headers.items():
                standard_headers[key] = value

        return self.return_funtion_response(
            HTMLResponse(
                content=default_page.render(
                    request=request,
                    root_element=[frame.render()],
                    title=self.__get_title__(frame, title),
                    theme=theme,
                    ext_js=ext_js,
                    ext_css=ext_css,
                    toast_delay=self.toast_delay,
                    error_classes=self.error_classes,
                    auth_header_name=self.auth_header,
                    description_content=frame.meta_description_content,
                    overflow=root_overflow,
                    head=frame.head_ext,
                    version=__version__,
                ),
                status_code=status_code,
                headers=standard_headers,
                media_type=media_type,
            )
        )

    def return_funtion_response(self, response: Union[str, Response]) -> Union[str, Response]:
        Frame.get_stack().del_stack()
        return response



from uiwiz.elements.extensions.on_event import OnEvent


class Toggle(OnEvent):
    root_class: str = "toggle"
    root_size: str = "toggle-{size}"

    def __init__(self, name: str, checked: bool = False) -> None:
        super().__init__("input")
        if checked:
            self.attributes["checked"] = "checked"
        self.attributes["type"] = "checkbox"
        self.attributes["name"] = name



import functools
import inspect
from typing import Callable, Optional, Union

from fastapi import APIRouter, Request, Response
from fastapi.responses import HTMLResponse

from uiwiz.asgi_request_middleware import get_request
from uiwiz.element import Element
from uiwiz.frame import Frame
from uiwiz.shared import register_path, route_exists
from uiwiz.version import __version__


class PageRouter(APIRouter):
    def page(
        self,
        path: str,
        *args,
        title: Optional[str] = None,
        favicon: Optional[str] = None,
        router: Optional[APIRouter] = None,
        **kwargs,
    ) -> Callable:
        def decorator(func: Callable, *args, **kwargs) -> Callable:
            parameters_of_decorated_func = list(inspect.signature(func).parameters.keys())

            @functools.wraps(func)
            async def decorated(*dec_args, **dec_kwargs) -> Response:
                Frame.get_stack().del_stack()
                # Create frame before function is called
                Frame.get_stack()
                Element().classes("flex flex-col min-h-screen h-full")
                request = dec_kwargs["request"]
                response = dec_kwargs["response"]

                # NOTE Ensure the signature matches the parameters of the function
                dec_kwargs = {k: v for k, v in dec_kwargs.items() if k in parameters_of_decorated_func}
                result = func(*dec_args, **dec_kwargs)
                if inspect.isawaitable(result):
                    result = await result

                if isinstance(result, Response):
                    return self.return_function_response(result)

                _app = get_request().app
                return _app.render(request, response, title)

            self.__ensure_request_response_signature__(decorated)

            _router = router or self

            return _router.get(path, *args, include_in_schema=False, **kwargs)(decorated)

        return decorator

    def ui(
        self,
        path: str,
        include_js: bool = True,
        include_css: bool = True,
        router: Optional[APIRouter] = None,
        **kwargs,
    ) -> Callable:
        def decorator(func: Callable) -> Callable:
            parameters_of_decorated_func = list(inspect.signature(func).parameters.keys())

            @functools.wraps(func)
            async def decorated(*dec_args, **dec_kwargs) -> Response:
                Frame.get_stack().del_stack()
                # Create frame before function is called
                Frame.get_stack()
                response = dec_kwargs["response"]

                # NOTE Ensure the signature matches the parameters of the function
                dec_kwargs = {k: v for k, v in dec_kwargs.items() if k in parameters_of_decorated_func}
                result = func(*dec_args, **dec_kwargs)
                if inspect.isawaitable(result):
                    result = await result

                if isinstance(result, Response):  # NOTE if setup returns a response, we don't need to render the page
                    return self.return_function_response(result)

                standard_headers = {"cache-control": "no-store", "x-uiwiz-content": "partial-ui"}
                for key, value in response.headers.items():
                    standard_headers[key] = value

                render = [Frame.get_stack().render()]
                js, css = Frame.get_stack().render_ext()
                if include_css:
                    render.append(css)
                if include_js:
                    render.append(js)
                content = "".join(render)

                return self.return_function_response(HTMLResponse(content=content, headers=standard_headers))

            self.__ensure_request_response_signature__(decorated)

            _router = router or self

            return _router.post(path, include_in_schema=False, **kwargs)(decorated)

        return decorator

    def return_function_response(self, response: Union[str, Response]) -> Union[str, Response]:
        Frame.get_stack().del_stack()
        return response

    def __ensure_request_response_signature__(self, func: Callable):
        data = {"request": Request, "response": Response}

        params = [p for p in inspect.signature(func).parameters.values()]
        for key, value in data.items():
            if key not in {p.name for p in params}:
                parm = inspect.Parameter(key, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=value)
                params.insert(0, parm)

        func.__signature__ = inspect.Signature(params)



from typing import Optional

from uiwiz.element import Element
from uiwiz.elements.extensions.bindable import Bindable


class Label(Bindable):
    root_class: str = "label "
    root_size: str = "label-{size}"

    def __init__(self, text: Optional[str] = None, for_: Optional[Element] = None) -> None:
        super().__init__(tag="label")
        if text:
            self.content = text
        if for_:
            self.attributes["for"] = for_.id

        # Allign the label next to the element
        self.attributes["style"] = "padding-top: unset;"

    def set_text(self, text: str) -> "Label":
        self.content = text
        return self

    def set_for(self, for_: Element) -> "Label":
        self.attributes["for"] = for_.id
        return self



from typing import TYPE_CHECKING, Callable, Literal, Optional, TypedDict, Union

if TYPE_CHECKING:
    from uiwiz.element import Element

ON_EVENTS = Literal[
    "input",
    "change",
    "click",
    "focus",
    "blur",
    "submit",
    "reset",
    "keydown",
    "keyup",
    "keypress",
    "scroll",
    "copy",
    "cut",
    "paste",
]

# HTMX Swap Events
# https://htmx.org/attributes/hx-swap/
SWAP_EVENTS = Optional[
    Literal[
        "innerHTML",
        "outerHTML",
        "textContent",
        "beforebegin",
        "afterbegin",
        "beforeend",
        "afterend",
        "delete",
        "none",
    ]
]

TARGET_TYPE = Optional[Union[Callable[[], str], str, "Element", None]]
FUNC_TYPE = Union[Callable, str]


class Event(TypedDict):
    func: FUNC_TYPE
    trigger: ON_EVENTS
    target: TARGET_TYPE
    swap: SWAP_EVENTS
    include: Optional[str]
    vals: Optional[str]
    params: Optional[dict]



from datetime import date, datetime
from typing import Callable, Optional, Union

from uiwiz.element import Element
from uiwiz.elements.extensions.on_event import OnEvent
from uiwiz.event import ON_EVENTS


class Datepicker(OnEvent):
    def __init__(
        self,
        name: str,
        value: Optional[datetime] = None,
    ) -> None:
        super().__init__("input")
        self.attributes["name"] = name
        self.attributes["type"] = "date"
        if value:
            self.default_date(value)

    def default_date(self, date: Union[datetime, date]) -> "Datepicker":
        if date is None:
            raise ValueError("Date cannot be None")
        self.value = date.strftime("%Y-%m-%d")
        return self



import numbers
from pathlib import Path
from typing import Iterable, Union

from uiwiz.element import Element


class Dict(Element):
    def __init__(self, data: Union[Iterable[dict], dict]) -> None:
        if not data:
            raise ValueError("Data cannot be None or empty")
        if not isinstance(data, (Iterable, dict)):
            raise ValueError("Data not of type list or dict")
        super().__init__()
        self.data = data
        self.did_render = False
        self.key_class = ""
        self.value_class = "text-primary"
        self._border_classes = "border border-base-content rounded-lg shadow-lg w-96 shadow-md w-full mb-5"

    def key_classes(self, classes):
        self.key_class = classes
        return self

    def value_classes(self, classes):
        self.value_class = classes
        return self

    def border_classes(self, classes: str):
        self._border_classes = classes
        return self

    def before_render(self):
        super().before_render()
        if not self.did_render:
            self.did_render = True
            self.generate(self.data)

    def generate(self, data):
        def format_data(
            data: Union[dict, list],
            depth: int = 0,
            is_last_item: bool = False,
            do_indent: bool = False,
            obj: bool = False,
        ):
            indent = " " * depth

            if isinstance(data, list):
                last_item = data[-1]
                Element("pre", content=indent + "[")
                for item in data:
                    is_last = item == last_item
                    with Element().classes("flex flex-col flex-wrap"):
                        format_data(item, depth=depth + 2, is_last_item=is_last, do_indent=True)
                Element("pre", content=indent + "]")
                return
            if isinstance(data, dict):
                last_item = list(data.values())[-1]
                is_last = data == last_item

                if not obj:
                    Element("pre", content=indent + "{")
                    format_data(data, depth=depth + 2, is_last_item=is_last, obj=True)
                else:
                    for key, value in data.items():
                        is_last = last_item == value
                        key_content = indent + f'"{key}"' + ":"
                        if isinstance(value, list):
                            key_content += " ["
                            with Element(tag="pre", content=key_content).classes("flex flex-col flex-wrap"):
                                for _item in value:
                                    format_data(_item, depth=depth + 2, is_last_item=_item == value[-1], do_indent=True)
                            Element(tag="pre", content=indent + "]" + ("," if not is_last else ""))
                        elif isinstance(value, dict):
                            key_content += " {"
                            with Element(tag="pre", content=key_content).classes("flex flex-col flex-wrap"):
                                format_data(value, depth=depth + 2, is_last_item=is_last, obj=True)
                            Element(tag="pre", content=indent + "}" + ("," if not is_last else ""))
                        else:
                            with Element(tag="pre", content=key_content).classes("flex flex-row flex-wrap gap-2"):
                                format_data(value, depth=depth + 2, is_last_item=is_last)
                if not obj:
                    Element("pre", content=indent + "}" + ("," if not is_last_item else ""))

                return

            if isinstance(data, str):
                with Element().classes("flex flex-row"):
                    if do_indent:
                        Element(tag="pre", content=indent)
                    Element(content=f'"{data}"').classes(self.value_class)
                    Element(content="," if not is_last_item else "")

            if isinstance(data, numbers.Number):
                with Element().classes("flex flex-row"):
                    if do_indent:
                        Element(tag="pre", content=indent)
                    Element(content=str(data)).classes(self.value_class)
                    Element(tag="div", content="," if not is_last_item else "")

        with self.classes(self._border_classes):
            format_data(data, is_last_item=True)



from __future__ import annotations

import asyncio
import os
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple, Union
from uuid import uuid4

from uiwiz.asgi_request_middleware import get_request
from uiwiz.version import __version__

if TYPE_CHECKING:
    from uiwiz.element import Element


def get_task_id() -> int:
    try:
        return id(asyncio.current_task())
    except RuntimeError:
        return 0


class Frame:
    stacks: dict[int, "Frame"] = {}

    def __init__(self) -> None:
        self.root_element: Optional[Element] = None
        self.current_element: Optional[Element] = None
        self.oob_elements: list[Element] = []
        self.id_count: int = 0  # used for element id
        self.scripts: list[str] = []
        self.extensions: List[str] = []
        self.app = get_request().app
        self.last_id = None
        self.title: Optional[str] = None
        self.meta_description_content: str = ""
        self.head_ext: str = ""

    def get_id(self) -> str:
        headers = get_request().headers
        swap = headers.get("hx-swap")

        if swap is None:
            return f"a-{self.id_count}"

        target_id = self.last_id if self.last_id else headers.get("hx-target")
        if swap.lower() in ["outerhtml", "this"] and self.last_id != target_id:
            self.last_id = target_id
        else:
            self.last_id = "a" + str(uuid4())

        return self.last_id

    def remove_frame(self, element: "Element") -> None:
        if element is None:
            if self.root_element:
                self.root_element.stack = None
                self.remove_frame(self.root_element)
        else:
            element.stack = None
            for child in element.children:
                self.remove_frame(child)

    def render(self) -> str:
        if self.root_element:
            return self.root_element.render()
        if self.oob_elements:
            output = ""
            el: Element
            for el in self.oob_elements:
                output += el.render_top_level()
            return output
        return ""

    def render_ext(self) -> Tuple[str, str]:
        if self.root_element:
            return self.root_element.render_ext(self.extensions)
        return "", ""

    def add_extension(self, cls, extensions: Optional[Union[list[Path], Path]]) -> None:
        if extensions is None:
            return
        if not isinstance(extensions, list):
            extensions = [extensions]

        for extension in extensions:
            _, filename = os.path.split(extension)
            prefix = f"/_static/extension/{__version__}/{cls.__name__}/"
            endpoint = prefix + filename
            if extension not in self.extensions:
                self.extensions.append(endpoint)

    @classmethod
    def get_stack(cls) -> "Frame":
        _id = get_task_id()
        if _id not in cls.stacks:
            cls.stacks[_id] = Frame()
        return cls.stacks[_id]

    @classmethod
    def del_stack(cls) -> None:
        cls.get_stack().remove_frame(None)
        del cls.stacks[get_task_id()]

    @classmethod
    def set_title(cls, title: str) -> None:
        if isinstance(title, str) is False:
            raise Exception(f"Expected str got: {type(title)}")
        Frame.get_stack().title = title

    @classmethod
    def set_meta_description_content(cls, content: str) -> None:
        if isinstance(content, str) is False:
            raise Exception(f"Expected str got: {type(content)}")
        Frame.get_stack().meta_description_content = content



import html
import json
from pathlib import Path

from fastapi.responses import JSONResponse

from uiwiz.element import Element

LIB_PATH = Path(__file__).parent / "echart.min.js"
JS_PATH = Path(__file__).parent / "echart.js"


class EChart(Element, extensions=[LIB_PATH, JS_PATH]):
    name: str = "data-wz-echart"

    def __init__(self, options: dict, height: str = "h-80") -> None:
        with Element() as container:
            self.parent_element = container
            container.classes(f"flex justify-center relative overflow-hidden items-center {height}")
            super().__init__()
            self.attributes[EChart.name] = EChart.name
            self.attributes[f"{EChart.name}-options"] = html.escape(json.dumps(options))
            self.attributes["hx-ext"] = EChart.name
            self.classes("w-full h-full")

    def container_classes(self, input: str) -> "EChart":
        self.parent_element.classes(input)
        return self

    @staticmethod
    def response(data: dict, headers: dict[str, str] = {}) -> JSONResponse:
        _headers = {"HX-Trigger": json.dumps({"uiwizUpdateEChart": data})} | headers
        return JSONResponse(data, headers=_headers)



import inspect
from dataclasses import dataclass
from datetime import date, datetime
from typing import Annotated, Literal, Optional, Tuple, Union, get_args, get_origin, get_type_hints

from pydantic import BaseModel
from pydantic_core import PydanticUndefinedType

from uiwiz.element import Element
from uiwiz.element_types import ELEMENT_SIZE
from uiwiz.elements.button import Button
from uiwiz.elements.checkbox import Checkbox
from uiwiz.elements.datepicker import Datepicker
from uiwiz.elements.divider import Divider
from uiwiz.elements.dropdown import Dropdown
from uiwiz.elements.form import Form
from uiwiz.elements.hidden_input import HiddenInput
from uiwiz.elements.input import Input
from uiwiz.elements.label import Label
from uiwiz.elements.radio import Radio
from uiwiz.elements.textarea import TextArea
from uiwiz.elements.toggle import Toggle
from uiwiz.models.display import display_name


@dataclass
class UiAnno:
    type: Union[Input, HiddenInput, Toggle, Datepicker, Dropdown, TextArea, Checkbox] = None
    placeholder: Optional[str] = None
    classes: Optional[str] = None


switch = {
    str: Input,
    int: Input,
    float: Input,
    bool: Toggle,
    Literal: Dropdown,
    date: Datepicker,
    datetime: Datepicker,
}


class ModelForm:
    model: BaseModel
    compact: bool
    form: Form
    button: Button
    label_classes: str
    instance: Optional[BaseModel]

    def __init__(
        self,
        model: BaseModel,
        compact: bool = True,
        card_classes: str = "border border-base-content rounded-lg shadow-lg w-full",
        label_classes: str = "flex-auto w-52",
        size: ELEMENT_SIZE = "md",
        **kwargs,  # override fields with custom ui
    ):
        self.model = model
        self.instance = None

        if isinstance(model, BaseModel):
            self.model = model.__class__
            self.instance = model
        self.compact = compact
        self.label_classes = label_classes
        self.card_classes = card_classes
        self._size = size
        self.render_model(**kwargs)
        self.button.render_html = False

    def on_submit(self, *args, **kwargs) -> "ModelForm":
        self.button.render_html = True
        self.form.on_submit(*args, **kwargs)
        return self

    def render_model(self, **kwargs) -> Form:
        if issubclass(self.model, BaseModel) == False:
            raise ValueError("type must be a pydantic model")

        with Form().classes(self.card_classes) as form:
            hints = get_type_hints(self.model, include_extras=True)
            for key, field_type in hints.items():
                self.render_model_attributes(key, field_type, **kwargs)
            self.button = Button("Save")
        self.form = form

    def render_model_attributes(self, key, field_type, **kwargs) -> "ModelForm":
        args = get_args(field_type)
        annotated = Annotated == get_origin(field_type)
        if key in kwargs:
            self.render_key_override(args, key, **kwargs)
        else:
            self.render_type_hint_without_args(args, annotated, field_type, key)
            self.render_with_args_annotated(args, annotated, field_type, key)

    def render_key_override(self, args: Tuple, key: str, **kwargs) -> None:
        if key in kwargs:
            model_args: dict = kwargs[key]
            model = model_args.pop("ui")
            placeholder = str(model_args.pop("placeholder", key))
            if not issubclass(model, Element):
                raise ValueError("type must be type of Element")
            if "name" not in model_args:
                model_args["name"] = key

            _, arg = self.get_type_and_uianno(args)
            classes = None
            if arg:
                classes = arg.classes
            self.render_element(model, key, classes, placeholder=placeholder, **model_args)
        else:
            raise ValueError("key not found in kwargs. Unable to render")

    def render_type_hint_without_args(self, args: Tuple, annotated: bool, field_type, key) -> Element:
        if len(args) == 0:
            if annotated:
                self.render_element(switch.get(field_type), key, placeholder=key)

            self.render_element(switch.get(field_type), key, placeholder=key)

    def render_with_args_annotated(self, args: Tuple, annotated: bool, field_type: Tuple, key: str) -> Element:
        if len(args) > 0:
            if annotated:
                field_type, ele = self.get_type_and_uianno(args)
                if ele:
                    placeholder = ele.placeholder if ele.placeholder else key
                    if field_args := get_args(field_type):
                        self.render_element_radio(ele, key, field_args)
                    else:
                        self.render_element(ele.type, key, ele.classes, placeholder=placeholder)
            else:
                ele = switch.get(get_origin(field_type))
                if ele:
                    self.render_element_dropdown(field_type, key, self.model.model_fields[key].default)

    def get_type_and_uianno(self, args: Tuple) -> Optional[Tuple[type, Optional[UiAnno]]]:
        if args:
            field_type = args[0]
            for arg in args:
                if isinstance(arg, UiAnno):
                    return field_type, arg
            return field_type, None
        return None, None

    def render_element(
        self,
        ele: Element,
        key: str,
        classes: Optional[str] = None,
        field_arg: Optional[str] = None,
        **kwargs,
    ) -> None:
        kwargs = {**{"name": key}, **kwargs}
        placeholder = "placeholder"
        kwargs[placeholder] = display_name(kwargs[placeholder])
        compact = self.compact
        with Element().classes("flex flex-nowrap w-full"):
            ele_args = [item[0] for item in inspect.signature(ele.__init__).parameters.items()]
            compact = self.extend_kwargs(kwargs, ele_args, key, field_arg, ele)

            label: Optional[Label] = None
            if ele is not HiddenInput:
                label = Label(display_name(key))

                if compact and inspect.signature(ele).parameters.get(placeholder):
                    label.render_html = False
                else:
                    label.classes(self.label_classes)
                    label.render_html = True
            if not inspect.signature(ele).parameters.get(placeholder):
                kwargs.pop(placeholder)

            if self.instance and "placeholder" in ele_args:
                kwargs["placeholder"] = key
            el: Element = ele(**kwargs)
            el.size(self._size)
            if classes:
                el.classes(classes)
            if label:
                label.set_for(el)

    def extend_kwargs(self, kwargs: dict, ele_args: list[str], key: str, field_arg: str, ele: Element) -> bool:
        compact = self.compact
        if self.instance and "value" in ele_args:
            kwargs["value"] = getattr(self.instance, key)

        # radio button
        if self.instance and isinstance(ele, Radio):
            kwargs["checked"] = "checked"
            compact = False
        elif self.instance and isinstance(ele, Toggle):
            kwargs["checked"] = "checked"
            compact = False
        return compact

    def render_element_radio(
        self,
        ele: UiAnno,
        key: str,
        field_args: Tuple,
    ) -> None:
        Divider()
        for arg in field_args:
            self.render_element(Radio, key, ele.classes, placeholder=arg, field_arg=arg)
        Divider()

    def render_element_dropdown(self, field_class: type, key: str, placeholder: Optional[str]) -> None:
        if isinstance(placeholder, PydanticUndefinedType):
            placeholder = __display_name__(key)

        self.render_element(Dropdown, key, placeholder=placeholder, items=get_args(field_class))



import uiwiz.ui as ui
from uiwiz.app import UiwizApp
from uiwiz.element import Element
from uiwiz.page_route import PageRouter
from uiwiz.version import __version__

__all__ = ["ui", "UiwizApp", "Element", "PageRouter", "__version__"]



from typing import Optional

from uiwiz import ui
from uiwiz.asgi_request_middleware import get_request
from uiwiz.element import Element


class ThemeSelector(Element):
    def __init__(self, themes: Optional[list[str]] = None) -> None:
        super().__init__()
        self.render_html = False
        self.themes = [
            "light",
            "dark",
            "nord",
            "cupcake",
            "pastel",
            "bumblebee",
            "emerald",
            "corporate",
            "synthwave",
            "retro",
            "cyberpunk",
            "valentine",
            "halloween",
            "garden",
            "forest",
            "aqua",
            "lofi",
            "fantasy",
            "wireframe",
            "black",
            "luxury",
            "dracula",
            "cmyk",
            "autumn",
            "business",
            "acid",
            "lemonade",
            "night",
            "coffee",
            "winter",
            "dim",
            "sunset",
        ]
        if themes:
            self.themes = themes

        placeholder = "Theme"
        if theme := get_request().cookies.get("data-theme"):
            placeholder = theme

        self.theme_selector = ui.dropdown("theme-selector", self.themes, placeholder)
        self.theme_selector.classes("min-w-32")
        self.setup_listener()

    def setup_listener(self):
        self.script = f"""
function selectTheme(value) {{
    console.log(value);
    element = document.getElementById("html");
    element.setAttribute("data-theme", value);
    document.cookie = `data-theme=${{value}}; Path=/`; 
}}

document.getElementById("{self.theme_selector.id}").addEventListener('change', function() {{
    selectTheme(this.value);
}});
"""



from typing import Callable, Optional, Union

from uiwiz.element import Element
from uiwiz.event import FUNC_TYPE, SWAP_EVENTS, TARGET_TYPE


class Form(Element):
    root_class: str = "flex flex-col items-start gap-4 p-4 "

    def __init__(self) -> None:
        super().__init__("form")

    def on_submit(
        self,
        func: FUNC_TYPE = None,
        target: TARGET_TYPE = None,
        swap: SWAP_EVENTS = "none",
        params: Optional[dict[str, str]] = None,
    ):
        self.event = {"func": func, "trigger": "submit", "target": target, "swap": swap, "params": params}
        return self



from uiwiz.elements.col import Col


class Container(Col):
    root_class: str = "container mx-auto max-w-[960px]"

    def __init__(self) -> None:
        super().__init__()
        self.classes(super().root_class)



from typing import Optional

from uiwiz.element import Element
from uiwiz.event import FUNC_TYPE, ON_EVENTS, SWAP_EVENTS, TARGET_TYPE


class OnEvent(Element):
    def on(
        self,
        trigger: ON_EVENTS,
        func: FUNC_TYPE,
        target: TARGET_TYPE = None,
        swap: SWAP_EVENTS = None,
        params: Optional[dict[str, str]] = None,
    ) -> "OnEvent":
        """
        Register the type of event to listen for and the function to call when the event is triggered.

        :param trigger: The type of event to listen for. This can be any valid html event type.
        :param func: The function/endpoint to call when the event is triggered.
        :param target: The target element that will be swaped with the response from the server.
        :param swap: The type of swap to perform. Set to 'none' to disable swapping.
        :param params: The parameters to pass to the function/endpoint. This could be a dynamic value that the function needs to execute. /some/path/{id}

        :return: The current instance of the element.
        """
        self.event = {"func": func, "trigger": trigger, "target": target, "swap": swap, "params": params}
        return self
